import typing as t

from .._core import AulosObject
from ..note import BaseNote
from .schemas import TunerSchema


class Tuner[NOTE: BaseNote](AulosObject[TunerSchema]):
    Note: type[NOTE]
    _ratios: t.ClassVar[tuple[float, ...]]
    _root: float

    def __new__(cls, *args, **kwargs) -> t.Self:
        if cls is Tuner:
            raise TypeError("Tuner cannot be instantiated directly.")
        return super().__new__(cls)

    def __init__(self, root: float, **kwargs):
        super().__init__(**kwargs)
        self._root = root

    def __init_subclass__(
        cls,
        *,
        ratios: tuple[float, ...],
        reference_notenumber: int,
        note: type[NOTE],
        **kwargs,
    ):
        schema = TunerSchema(
            reference_notenumber,
            note.schema,
            note.schema.pitchclass,
        )
        super().__init_subclass__(schema=schema, **kwargs)
        cls.Note = note
        cls._ratios = ratios

    @property
    def root(self) -> float:
        return self._root

    def hz(self, notenumber: int) -> float:
        ref = notenumber - self.schema.reference_notenumber
        octnumber = ref // self.schema.pitchclass.cardinality
        pitchclass = ref % self.schema.pitchclass.cardinality
        return self._root * (2**octnumber) * self._ratios[pitchclass]

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Tuner):
            return NotImplemented
        return self._ratios == other._ratios

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return "<Tuner: {}>".format(self.__class__.__name__)

    def __repr__(self) -> str:
        return "<Tuner: {}>".format(self.__class__.__name__)



import typing as t

from .._core import AulosObject
from .pitchclass import BasePitchClass
from .schemas import KeySchema


class BaseKey[PITCHCLASS: BasePitchClass](AulosObject[KeySchema]):
    PitchClass: type[PITCHCLASS]
    _pitchname: str
    _pitchclass: int
    _signatures: tuple[int, ...]

    def __init__(self, identify: str | t.Self, **kwargs) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, BaseKey):
            self._pitchname = identify._pitchname
            self._pitchclass = identify._pitchclass
            self._signatures = identify._signatures

        elif self.is_keyname(identify):
            self._pitchname = identify
            self._pitchclass = self.schema.pitchclass.convert_pitchname_to_picthclass(
                identify
            )
            self._signatures = self.schema.generate_key_signatures(identify)

        else:
            raise ValueError()

    def __init_subclass__(
        cls, *, accidental: int, pitchclass: type[BasePitchClass], **kwargs
    ) -> None:
        schema = KeySchema(
            accidental,
            pitchclass.schema,
        )
        super().__init_subclass__(schema=schema, **kwargs)
        cls.PitchClass = pitchclass

    @property
    def keyname(self) -> str:
        return self._pitchname

    @property
    def signature(self) -> tuple[int, ...]:
        return self._signatures

    def to_pitchclass(self) -> PITCHCLASS:
        return self.PitchClass(self._pitchname, setting=self._setting)

    @classmethod
    def is_keyname(cls, value: t.Any) -> t.TypeGuard[str]:
        return cls.schema.is_keyname(value)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, t.SupportsInt):
            return NotImplemented
        return int(self) == int(other)

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __int__(self) -> int:
        return self._pitchclass

    def __str__(self) -> str:
        return f"<Key: {self.keyname}>"

    def __repr__(self) -> str:
        return f"<Key: {self.keyname}>"



import typing as t
from typing import TYPE_CHECKING

from .._core import AulosObject
from .._core.utils import index
from .pitchclass import BasePitchClass
from .schemas import NoteSchema

if TYPE_CHECKING:
    from ..scale import Scale  # pragma: no cover
    from ..tuner import Tuner  # pragma: no cover


class BaseNote[PITCHCLASS: BasePitchClass](AulosObject[NoteSchema]):
    PitchClass: type[PITCHCLASS]
    _notenumber: int
    _notenames: tuple[str | None, ...]
    _notename: str | None
    _tuner: t.Optional["Tuner"]
    _scale: t.Optional["Scale"]

    def __init__(
        self,
        identify: int | str,
        *,
        tuner: t.Optional["Tuner"] = None,
        scale: t.Optional["Scale"] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, BaseNote):
            self._notenumber = identify._notenumber
            self._notenames = identify._notenames
            self._notename = identify._notename
            self._scale = None
            self._tuner = identify._tuner
            self.scale = identify._scale

        elif self.is_notenumber(identify):
            notenames = self.schema.convert_notenumber_to_notenames(identify)
            self._notenumber = identify
            self._notenames = notenames
            self._notename = None
            self._scale = None
            self._tuner = tuner
            self.scale = scale

        elif self.is_notename(identify):
            notenumber = self.schema.convert_notename_to_notenumber(identify)
            notenames = self.schema.convert_notenumber_to_notenames(notenumber)
            self._notenumber = notenumber
            self._notenames = notenames
            self._notename = identify
            self._scale = None
            self._tuner = tuner
            self.scale = scale

        else:
            raise ValueError()

    def __init_subclass__(
        cls,
        *,
        symbols_notenumber: t.Sequence[int],
        symbols_octave: t.Sequence[str],
        reference_notenumber: int,
        reference_octave: int,
        pitchclass: type[PITCHCLASS],
        **kwargs,
    ) -> None:
        schema = NoteSchema(
            tuple(symbols_notenumber),
            tuple(symbols_octave),
            reference_notenumber,
            reference_octave,
            pitchclass.schema,
        )
        super().__init_subclass__(schema=schema, **kwargs)
        cls.PitchClass = pitchclass

    @property
    def notenumber(self) -> int:
        return self._notenumber

    @property
    def notenames(self) -> list[str]:
        return [n for n in self._notenames if n is not None]

    @property
    def notename(self) -> str | None:
        return self._notename

    @notename.setter
    def notename(self, name: str):
        if self.is_notename(name) and name in self._notenames:
            self._notename = name

    @property
    def tuner(self) -> t.Optional["Tuner"]:
        return self._tuner

    @tuner.setter
    def tuner(self, tuner: t.Optional["Tuner"]):
        from ..tuner import Tuner

        if isinstance(tuner, Tuner):
            self._tuner = tuner

    @property
    def scale(self) -> t.Optional["Scale"]:
        return self._scale

    @scale.setter
    def scale(self, scale: t.Optional["Scale"]):
        from ..scale import Scale

        if isinstance(scale, Scale):
            self._scale = scale
            pitchclass = self.schema.convert_notenumber_to_pitchclass(self._notenumber)
            pitchclass = (
                pitchclass - scale.key.pitchclass
            ) % self.schema.pitchclass.cardinality

            if (idx := index(scale.positions, pitchclass)) is not None:
                self._notename = self.schema.convert_notenumber_to_notename(
                    self._notenumber, scale.signatures[idx]
                )

    @property
    def hz(self) -> float | None:
        if self._tuner is None:
            return None
        return self._tuner.hz(self._notenumber)

    def to_pitchclass(self) -> PITCHCLASS:
        pitchlass = self.schema.convert_notenumber_to_pitchclass(self._notenumber)
        return self.PitchClass(
            pitchlass, tuner=self._tuner, scale=self._scale, setting=self._setting
        )

    @classmethod
    def is_notename(cls, notename: t.Any) -> t.TypeGuard[str]:
        return cls.schema.is_notename(notename)

    @classmethod
    def is_notenumber(cls, notenumber: t.Any) -> t.TypeGuard[int]:
        return cls.schema.is_notenumber(notenumber)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, t.SupportsInt):
            return NotImplemented
        return int(self) == int(other)

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __add__(self, other: t.SupportsInt) -> t.Self:
        return self.__class__(
            int(self) + int(other), scale=self.scale, setting=self.setting
        )

    def __sub__(self, other: t.SupportsInt) -> t.Self:
        return self.__class__(
            int(self) - int(other), scale=self.scale, setting=self.setting
        )

    def __int__(self):
        return self._notenumber

    def __str__(self) -> str:
        return f"<Note: {self.notename or self.notenames}, scale: {self.scale}>"

    def __repr__(self) -> str:
        return f"<Note: {self.notename or self.notenames}, scale: {self.scale}>"



import pytest

from src.aulos.TET12 import Note


def test_Note_init_from_notename(data_notenames):
    for pitchname in data_notenames:
        assert isinstance(Note(pitchname), Note)


def test_Note_init_from_notenumber(data_notenumbers):
    for pitchclass in data_notenumbers:
        assert isinstance(Note(pitchclass), Note)


@pytest.mark.parametrize(
    "invalid_value",
    ["", " ", "Cb-1", "G#9", -1, 128, None, [], {}],
)
def test_Note_init_from_invalid_value(invalid_value):
    with pytest.raises(ValueError):
        _ = Note(invalid_value)


def test_Note_property_get_notenumber(data_notenumbers):
    for notenumber in data_notenumbers:
        assert Note(notenumber).notenumber == notenumber


def test_Note_property_get_notename(data_notenumbers, data_notenames):
    for notenumber in data_notenumbers:
        assert Note(notenumber).notename is None
    for notename in data_notenames:
        assert Note(notename).notename == notename


def test_Note_property_get_notenames(
    data_notenumbers,
    data_notenames,
    data_map_notenumber_to_notenames,
    data_map_notename_to_notenumber,
):
    for notenumber in data_notenumbers:
        assert Note(notenumber).notenames == [
            item
            for item in data_map_notenumber_to_notenames[notenumber]
            if item is not None
        ]
    for notename in data_notenames:
        assert Note(notename).notenames == [
            item
            for item in data_map_notenumber_to_notenames[
                data_map_notename_to_notenumber[notename]
            ]
            if item is not None
        ]


def test_Note_dunder_eqne(data_map_notename_to_notenumber):
    for notename, notenumber in data_map_notename_to_notenumber.items():
        assert not Note(notename) != notenumber
        assert not Note(notenumber) != notenumber
        assert not Note(notename) != Note(notenumber)
        assert not Note(notenumber) != Note(notename)


def test_Note_dunder_add(data_notenumbers):
    for notenumber1 in data_notenumbers:
        for notenumber2 in data_notenumbers:
            if (notenumber1 + notenumber2) in data_notenumbers:
                assert (Note(notenumber1) + notenumber2) == (notenumber1 + notenumber2)
            else:
                with pytest.raises(ValueError):
                    _ = Note(notenumber1) + notenumber2


def test_Note_dunder_sub(data_notenumbers):
    for notenumber1 in data_notenumbers:
        for notenumber2 in data_notenumbers:
            if (notenumber1 - notenumber2) in data_notenumbers:
                assert (Note(notenumber1) - notenumber2) == (notenumber1 - notenumber2)
            else:
                with pytest.raises(ValueError):
                    _ = Note(notenumber1) - notenumber2


def test_Note_dunder_int(data_notenumbers):
    for notenumber in data_notenumbers:
        assert int(Note(notenumber)) == notenumber


def test_PitchClass_dunder_str(
    data_notenumbers, data_notenames, data_map_notenumber_to_notenames
):
    for notenumber in data_notenumbers:
        notenames = [
            name
            for name in data_map_notenumber_to_notenames[notenumber]
            if name is not None
        ]
        assert str(Note(notenumber)) == f"<Note: {notenames}, scale: None>"
    for notename in data_notenames:
        assert str(Note(notename)) == f"<Note: {notename}, scale: None>"



"""aulos (library for music theory)"""

from . import TET12, TET24
from ._core import Setting
from ._errors import *  # noqa: F403
from ._warnings import *  # noqa: F403
from .note import BaseKey, BaseNote, BasePitchClass
from .scale import DiatonicScale, NondiatonicScale, Scale
from .tuner import Tuner

__all__ = [
    "BasePitchClass",
    "BaseKey",
    "BaseNote",
    "Scale",
    "DiatonicScale",
    "NondiatonicScale",
    "Tuner",
    "Setting",
    "TET12",
    "TET24",
]



import typing as t
from collections import deque


def index[T](sequence: t.Sequence[T], target: T) -> int | None:
    if target not in sequence:
        return None
    return sequence.index(target)


def rotated[T](sequence: t.Sequence[T], shift: int = 0) -> tuple[T, ...]:
    rotated = deque(sequence)
    rotated.rotate(shift)
    return tuple(rotated)


def inplace[T](sequence: t.MutableSequence[T], f: t.Callable[[T], T]):
    for i in range(len(sequence)):
        sequence[i] = f(sequence[i])



import json
import os
import typing as t
from dataclasses import dataclass
from pathlib import Path

import tomllib

from .utils import from_dict


@dataclass(frozen=True, slots=True)
class Setting:
    @classmethod
    def default(cls) -> t.Self:
        path = Path(os.path.dirname(__file__)) / "default.toml"
        return cls.from_toml(path)

    @classmethod
    def from_dict(cls, value: dict[str, t.Any]) -> t.Self:
        return from_dict(cls, value)

    @classmethod
    def from_toml(cls, path: Path) -> t.Self:
        setting = tomllib.load(open(path, mode="rb"))
        return from_dict(cls, setting)

    @classmethod
    def from_json(cls, path: Path) -> t.Self:
        setting = json.load(open(path, mode="rb"))
        return from_dict(cls, setting)



"""Core
---
"""

# utils, framework
from . import framework, utils

# implementation
from .object import AulosObject
from .schema import Schema
from .setting import Setting

__all__ = [
    "AulosObject",
    "Setting",
    "Schema",
    "framework",
    "utils",
]



import typing as t
from itertools import pairwise
from typing import TYPE_CHECKING

from .._core.utils import classproperty, rotated
from .scale import Scale

# type annotaion
if TYPE_CHECKING:
    from ..note import (
        BaseKey,  # pragma: no cover
        BasePitchClass,  # pragma: no cover
    )


class DiatonicScale[KEY: BaseKey, PITCHCLASS: BasePitchClass](Scale[KEY, PITCHCLASS]):
    def __new__(cls, *args, **kwargs) -> t.Self:
        if cls is DiatonicScale:
            raise TypeError("DiatonicScale cannot be instantiated directly.")
        return super().__new__(cls)

    def __init_subclass__(
        cls,
        *,
        intervals: t.Sequence[int],
        shift: int = 0,
        key: type[KEY],
        pitchclass: type[PITCHCLASS],
        **kwargs,
    ) -> None:
        super().__init_subclass__(
            intervals=rotated(intervals, -shift),
            key=key,
            pitchclass=pitchclass,
            **kwargs,
        )


class NondiatonicScale[KEY: BaseKey, PITCHCLASS: BasePitchClass](
    Scale[KEY, PITCHCLASS]
):
    _extensions: t.ClassVar[tuple[tuple[int, ...], ...]]
    _base: t.ClassVar[type[Scale]]

    def __new__(cls, *args, **kwargs) -> t.Self:
        if cls is NondiatonicScale:
            raise TypeError("NondiatonicScale cannot be instantiated directly.")
        return super().__new__(cls)

    def __init_subclass__(
        cls,
        *,
        extensions: t.Sequence[t.Sequence[int]],
        base: type[DiatonicScale],
        key: type[KEY],
        pitchclass: type[PITCHCLASS],
        **kwargs,
    ) -> None:
        super().__init_subclass__(
            intervals=base.intervals,
            key=key,
            pitchclass=pitchclass,
            **kwargs,
        )
        cls._base = base
        cls._extensions = tuple(tuple(inner) for inner in extensions)

    @classproperty
    def intervals(cls) -> tuple[int, ...]:
        intervals = tuple(
            b - a for a, b in pairwise(cls.positions + (sum(super().intervals),))
        )
        return intervals

    @classproperty
    def positions(cls) -> tuple[int, ...]:
        positions = tuple(
            pos + ext
            for pos, exts in zip(super().positions, cls._extensions)
            for ext in exts
        )
        return positions

    @property
    def signatures(self) -> tuple[int, ...]:
        signatures = tuple(
            sig + ext
            for sig, exts in zip(super().signatures, self._extensions)
            for ext in exts
        )
        return signatures



import pytest

from src.aulos.TET12 import PitchClass


def test_PitchClass_init_from_pitchname(data_pitchname):
    for pitchname in data_pitchname:
        assert isinstance(PitchClass(pitchname), PitchClass)


def test_PitchClass_init_from_pitchclass(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert isinstance(PitchClass(pitchclass), PitchClass)


@pytest.mark.parametrize(
    "invalid_value",
    ["", " ", "a", "g", "@", "H", -1, 12, None, [], {}],
)
def test_PitchClass_init_from_invalid_value(invalid_value):
    with pytest.raises(ValueError):
        _ = PitchClass(invalid_value)


def test_PitchClass_property_get_pitchclass(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert PitchClass(pitchclass).pitchclass == pitchclass


def test_PitchClass_property_get_pitchname(data_pitchclass, data_pitchname):
    for pitchclass in data_pitchclass:
        assert PitchClass(pitchclass).pitchname is None
    for pitchname in data_pitchname:
        assert PitchClass(pitchname).pitchname == pitchname


def test_PitchClass_property_get_pitchnames(
    data_pitchclass,
    data_pitchname,
    data_map_pitchclass_to_pitchnames,
    data_map_pitchname_to_pitchclass,
):
    for pitchclass in data_pitchclass:
        assert PitchClass(pitchclass).pitchnames == [
            item
            for item in data_map_pitchclass_to_pitchnames[pitchclass]
            if item is not None
        ]
    for pitchname in data_pitchname:
        assert PitchClass(pitchname).pitchnames == [
            item
            for item in data_map_pitchclass_to_pitchnames[
                data_map_pitchname_to_pitchclass[pitchname]
            ]
            if item is not None
        ]


def test_PitchClass_dunder_eqne(data_map_pitchname_to_pitchclass):
    for pitchname, pitchclass in data_map_pitchname_to_pitchclass.items():
        assert not PitchClass(pitchname) != pitchclass
        assert not PitchClass(pitchclass) != pitchclass
        assert not PitchClass(pitchname) != PitchClass(pitchclass)
        assert not PitchClass(pitchclass) != PitchClass(pitchname)


def test_PitchClass_dunder_add(data_pitchclass):
    for pitchclass in data_pitchclass:
        for pitchclass2 in data_pitchclass:
            assert (PitchClass(pitchclass) + pitchclass2) == (
                pitchclass + pitchclass2
            ) % sum(PitchClass.schema.intervals)


def test_PitchClass_dunder_sub(data_pitchclass):
    for pitchclass in data_pitchclass:
        for pitchclass2 in data_pitchclass:
            assert (PitchClass(pitchclass) - pitchclass2) == (
                pitchclass - pitchclass2
            ) % sum(PitchClass.schema.intervals)


def test_PitchClass_dunder_int(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert int(PitchClass(pitchclass)) == pitchclass


def test_PitchClass_dunder_str(
    data_pitchclass, data_pitchname, data_map_pitchclass_to_pitchnames
):
    for pitchclass in data_pitchclass:
        pitchnames = [
            name
            for name in data_map_pitchclass_to_pitchnames[pitchclass]
            if name is not None
        ]
        assert str(PitchClass(pitchclass)) == f"<PitchClass: {pitchnames}, scale: None>"
    for pitchname in data_pitchname:
        assert str(PitchClass(pitchname)) == f"<PitchClass: {pitchname}, scale: None>"



import tkinter as tk
from abc import ABCMeta, abstractmethod


class BaseComponent(tk.Frame, metaclass=ABCMeta):
    def __init__(self, master: tk.Misc, **kwargs):
        super().__init__(master, padx=12, pady=4, **kwargs)

    @abstractmethod
    def create_widget(self, *args, **kwargs) -> None: ...

    @abstractmethod
    def default(self) -> None: ...



import typing as t
from abc import ABCMeta

from .framework import InjectedMeta, OptimizedMeta
from .schema import Schema
from .setting import Setting
from .utils import classproperty


class AulosObjectMeta(InjectedMeta, OptimizedMeta, ABCMeta):
    """This metaclass enables dependency injection, optimizations, and abstract base class capabilities."""


class AulosObject[T: Schema, *_](metaclass=AulosObjectMeta):
    _schema: T
    _setting: Setting | None

    def __new__(cls, *args, **kwargs) -> t.Self:
        if cls is AulosObject:
            raise TypeError("AulosObject cannot be instantiated directly.")
        return super().__new__(cls)

    def __init__(self, setting: Setting | None = None) -> None:
        super(AulosObject, self).__init__()
        self._setting = setting

    def __init_subclass__(cls, *, schema: T | None = None) -> None:
        if schema is None:
            return
        super(AulosObject, cls).__init_subclass__()
        cls._schema = schema

    @classproperty
    def schema(cls) -> T:
        return cls._schema

    @property
    def setting(self) -> Setting | None:
        return self._setting

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, AulosObject):
            return NotImplemented
        return self._schema == self._schema and self._setting == other._setting

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return "<AulosObject: setting={}>".format(self._setting)

    def __repr__(self) -> str:
        return "<AulosObject: setting={}>".format(self._setting)



import tkinter as tk
import tkinter.ttk as ttk
import typing as t

from ...const import KEY_DEFAULTS
from ..base import BaseComponent


class KeySelecter(BaseComponent):
    _selected_keyname: tk.StringVar

    _keyselecter_wrap: ttk.Frame
    _keyselecter_title: ttk.Label
    _keygroups: list[ttk.Frame]
    _keybuttons: list[list[ttk.Radiobutton]]

    def __init__(self, master: tk.Misc):
        super().__init__(master)
        self.master = master
        self.create_widget()

    def create_widget(self):
        self._selected_keyname = tk.StringVar()
        self._keyselecter_wrap = ttk.Frame(
            self, padding=(24, 8), borderwidth=2, relief=tk.SOLID
        )
        self._keyselecter_title = ttk.Label(self, text="Key")
        self._keyselecter_wrap.pack()
        self._keyselecter_title.place(relx=0.05, rely=0, anchor=tk.W)

        self._keygroups = [
            ttk.Frame(self._keyselecter_wrap, padding=(6, 0))
            for _ in range(len(KEY_DEFAULTS))
        ]
        self._keybuttons = [
            [
                ttk.Radiobutton(
                    keygroup,
                    text=key,
                    value=key,
                    variable=self._selected_keyname,
                    command=self._onClickKeyButton,
                )
                for key in keys
            ]
            for keygroup, keys in zip(self._keygroups, KEY_DEFAULTS)
        ]

        for keygroup in self._keygroups:
            keygroup.pack(side=tk.LEFT, anchor=tk.NW)

        for keybuttons in self._keybuttons:
            for btn in keybuttons:
                btn.pack(side=tk.TOP, anchor=tk.NW)

    def default(self):
        self._selected_keyname.set("C")
        self._onClickKeyButton()

    def _onClickKeyButton(self):
        for callback in self.callbacks_onClickKeyButton:
            callback()

    @property
    def keyname(self):
        return self._selected_keyname.get()

    callbacks_onClickKeyButton: list[t.Callable[[], t.Any]]

    def set_callback_onClickKeyButton(self, callback: t.Callable[[], t.Any]):
        if not hasattr(self, "callbacks_onClickKeyButton"):
            self.callbacks_onClickKeyButton = []
        self.callbacks_onClickKeyButton.append(callback)



import typing as t
from dataclasses import dataclass, field
from functools import cached_property
from itertools import accumulate, chain

from ..._core import Schema


@dataclass(frozen=True, slots=True)
class PitchClassSchema(Schema):
    intervals: tuple[int, ...]
    symbols_pitchclass: tuple[str, ...]
    symbols_accidental: tuple[str, ...]

    cardinality: int = field(init=False)
    accidental: int = field(init=False)
    positions: tuple[int, ...] = field(init=False)
    name2class: dict[str, int] = field(init=False)
    class2name: dict[int, tuple[str | None, ...]] = field(init=False)

    def __post_init__(self) -> None:
        self.validate()
        self.initialize()

    def validate(self) -> None:
        # [check] intervals
        if not len(self.intervals) > 0:
            raise Exception()
        if not all(0 <= v for v in self.intervals):
            raise Exception()

        # [check] symbols_pitchclass
        if not len(self.symbols_pitchclass) > 0:
            raise Exception()

        # [check] symbols_accidental
        if not len(self.symbols_accidental) > 0:
            raise Exception()
        if not (len(self.symbols_accidental) % 2) == 0:
            raise Exception()

        # [cross-field check]
        if not len(self.intervals) == len(self.symbols_pitchclass):
            raise Exception()

    def initialize(self) -> None:
        cardinality = sum(self.intervals)
        positions = tuple(accumulate((0,) + self.intervals[:-1]))

        accidental = len(self.symbols_accidental) // 2
        upper_accidentals = self.symbols_accidental[accidental:]
        lower_accidentals = reversed(self.symbols_accidental[:accidental])

        def create_upper_sequences() -> list[list[str | None]]:
            sequences = []
            for i, acc in enumerate(upper_accidentals, start=1):
                sequence = create_symbol_sequence(suffix=acc)
                for _ in range(i):
                    sequence.insert(0, sequence.pop())
                sequences.append(sequence)
            return sequences

        def create_lower_sequences() -> list[list[str | None]]:
            sequences = []
            for i, acc in enumerate(lower_accidentals, start=1):
                sequence = create_symbol_sequence(suffix=acc)
                for _ in range(i):
                    sequence.append(sequence.pop(0))
                sequences.append(sequence)
            return sequences

        def create_symbol_sequence(
            *, prefix: str = "", suffix: str = ""
        ) -> list[str | None]:
            sequence: list[str | None] = []
            for deg in range(cardinality):
                if deg in positions:
                    index = positions.index(deg)
                    sequence.append(prefix + self.symbols_pitchclass[index] + suffix)
                else:
                    sequence.append(None)
            return sequence

        no_accidental_sequence = create_symbol_sequence()
        accidental_upper_sequences = create_upper_sequences()
        accidental_lower_sequences = reversed(create_lower_sequences())
        accidental_sequences = tuple(
            zip(
                *accidental_lower_sequences,
                no_accidental_sequence,
                *accidental_upper_sequences,
            )
        )
        name2class = [
            [(name, index) for name in names if name is not None]
            for index, names in enumerate(accidental_sequences)
        ]
        class2name = [(index, name) for index, name in enumerate(accidental_sequences)]

        object.__setattr__(self, "cardinality", cardinality)
        object.__setattr__(self, "accidental", accidental)
        object.__setattr__(self, "positions", positions)
        object.__setattr__(self, "name2class", dict(chain.from_iterable(name2class)))
        object.__setattr__(self, "class2name", dict(class2name))

    @cached_property
    def pitchnames(self) -> tuple[str, ...]:
        return tuple(self.name2class.keys())

    @cached_property
    def pitchclasses(self) -> tuple[int, ...]:
        return tuple(self.class2name.keys())

    # [unstable]
    def find_pitchname(self, value: str) -> str | None:
        finded = sorted(
            [pitchname for pitchname in self.pitchnames if value.find(pitchname) == 0],
            key=len,
            reverse=True,
        )
        return (finded + [None])[0]

    def count_accidental(self, pitchname: str) -> int:
        self.ensure_valid_pitchname(pitchname)
        pitchclass = self.convert_pitchname_to_picthclass(pitchname)
        pitchnames = self.convert_pitchclass_to_pitchnames(pitchclass)
        return pitchnames.index(pitchname) - self.accidental

    def convert_pitchclass_to_pitchname(
        self, pitchclass: int, accidental: int
    ) -> str | None:
        self.ensure_valid_pitchclass(pitchclass)
        self.ensure_valid_accidental(accidental)
        return self.class2name[pitchclass][self.accidental + accidental]

    def convert_pitchclass_to_pitchnames(
        self, pitchclass: int
    ) -> tuple[str | None, ...]:
        self.ensure_valid_pitchclass(pitchclass)
        return self.class2name[pitchclass]

    def convert_pitchname_to_picthclass(self, pitchname: str) -> int:
        self.ensure_valid_pitchname(pitchname)
        return self.name2class[pitchname]

    def convert_pitchclass_to_symbol(self, pitchclass: int) -> str | None:
        self.ensure_valid_pitchclass(pitchclass)
        return self.convert_pitchclass_to_pitchnames(pitchclass)[self.accidental]

    def convert_pitchname_to_symbol(self, pitchname: str) -> str:
        self.ensure_valid_pitchname(pitchname)
        accidental = self.count_accidental(pitchname)
        pitchclass = self.convert_pitchname_to_picthclass(pitchname)
        pitchclass = (pitchclass - accidental) % self.cardinality
        symbol = self.convert_pitchclass_to_pitchname(pitchclass, 0)
        if symbol is None:
            raise RuntimeError("unreachable error")
        return symbol

    def is_symbol(self, value: t.Any) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.symbols_pitchclass

    def is_pitchname(self, value: t.Any) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.pitchnames

    def is_pitchclass(self, value: t.Any) -> t.TypeGuard[int]:
        return isinstance(value, int) and value in self.pitchclasses

    def ensure_valid_pitchname(self, pitchname: str) -> None:
        if not self.is_pitchname(pitchname):
            raise ValueError(
                f"Invalid pitchname '{pitchname}'. "
                f"Pitchname must be a valid musical note name {self.pitchnames[:3]}."
            )

    def ensure_valid_pitchclass(self, pitchclass: int) -> None:
        if not self.is_pitchclass(pitchclass):
            raise ValueError(
                f"Invalid pitchclass '{pitchclass}'."
                f"Pitchclass must be an integer between {min(self.pitchclasses)} and {max(self.pitchclasses)} inclusive."
            )

    def ensure_valid_accidental(self, accidental: int) -> None:
        if not abs(accidental) <= self.accidental:
            raise ValueError(
                f"Invalid accidental '{accidental}'. "
                f"Accidental must be within the range -{self.accidental} to +{self.accidental}."
            )



"""Note
---
"""

from . import schemas
from .key import BaseKey
from .note import BaseNote
from .pitchclass import BasePitchClass

__all__ = [
    "BasePitchClass",
    "BaseKey",
    "BaseNote",
    "schemas",
]



import typing as t
from itertools import accumulate, starmap

from .._core import AulosObject
from .._core.utils import classproperty
from ..note import BaseKey, BasePitchClass
from .schemas import ScaleSchema


class Scale[KEY: BaseKey, PITCHCLASS: BasePitchClass](AulosObject[ScaleSchema]):
    Key: type[KEY]
    PitchClass: type[PITCHCLASS]
    _intervals: t.ClassVar[tuple[int, ...]]
    _positions: t.ClassVar[tuple[int, ...]]
    _key: KEY
    _signatures: tuple[int, ...]

    def __new__(cls, *args, **kwargs) -> t.Self:
        if cls is Scale:
            raise TypeError("Scale cannot be instantiated directly.")
        return super().__new__(cls)

    def __init__(self, key: str | KEY, **kwargs) -> None:
        super().__init__(**kwargs)

        if isinstance(key, str):
            self._key = self.Key(key, setting=self._setting)
            self._signatures = tuple(
                starmap(
                    lambda x, y: x + y,
                    zip(
                        self._key.signature,
                        self.schema.generate_scale_signatures(self._intervals),
                    ),
                )
            )

        elif isinstance(key, BaseKey):
            self._key = key
            self._signatures = tuple(
                starmap(
                    lambda x, y: x + y,
                    zip(
                        self._key.signature,
                        self.schema.generate_scale_signatures(self._intervals),
                    ),
                )
            )

        else:
            raise ValueError()

    def __init_subclass__(
        cls,
        *,
        intervals: t.Sequence[int] | None = None,
        key: type[KEY] | None = None,
        pitchclass: type[PITCHCLASS] | None = None,
        **kwargs,
    ) -> None:
        if intervals is None or key is None or pitchclass is None:
            return None
        schema = ScaleSchema(pitchclass.schema)
        super().__init_subclass__(schema=schema, **kwargs)
        cls.Key = key
        cls.PitchClass = pitchclass
        cls._intervals = tuple(intervals)
        cls._positions = tuple(accumulate((0,) + cls._intervals[:-1]))

    @property
    def key(self) -> KEY:
        return self._key

    @classproperty
    def intervals(cls) -> tuple[int, ...]:
        return cls._intervals

    @classproperty
    def positions(cls) -> tuple[int, ...]:
        return cls._positions

    @property
    def signatures(self) -> tuple[int, ...]:
        return self._signatures

    @property
    def components(self) -> tuple[PITCHCLASS, ...]:
        components = []
        root = self.PitchClass(self._key.keyname, scale=self, setting=self.setting)
        for pos in self.positions:
            pitchclass = (root + pos).pitchclass
            note = self.PitchClass(pitchclass, scale=self, setting=self.setting)
            components.append(note)
        return tuple(components)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Scale):
            return NotImplemented
        return self._intervals == other._intervals and self._key == other._key

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}: {self._key}>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._key}>"



import tkinter as tk
import tkinter.ttk as ttk

from ..const import KEYBOARD_BLACK_CLASSES, KEYBOARD_WHITE_CLASSES


class KeyElement(tk.Frame):
    key: tk.Button
    is_active: tk.BooleanVar

    notenumber: int

    def __init__(self, master: tk.Misc, notenumber: int, *, width: int, height: int):
        super().__init__(master)
        self.master = master
        self.notenumber = notenumber
        self.create_widget(width, height)

    def create_widget(self, width: int, height: int):
        if self.is_white():
            self.key = tk.Button(self)
            self.key.config(bg="#FFFFFF")
            self.config(width=width, height=height)
            self.key.place(x=0, y=0, anchor=tk.NW, width=width, height=height)

        elif self.is_black():
            self.key = tk.Button(self)
            self.key.config(bg="#000000")
            self.config(width=width * (2 / 3), height=height * (2 / 3))
            self.key.place(
                x=0, y=0, anchor=tk.NW, width=width * (2 / 3), height=height * (2 / 3)
            )

        def callback_btn_bgcolor(*args):
            if self.is_active.get():
                self.key.config(bg="#FF6347")
            else:
                if self.is_white():
                    self.key.config(bg="#FFFFFF")
                elif self.is_black():
                    self.key.config(bg="#000000")

        self.is_active = tk.BooleanVar(value=True)
        self.is_active.trace_add("write", callback_btn_bgcolor)

    def default(self):
        return

    def is_white(self) -> bool:
        return (self.notenumber % 12) in KEYBOARD_WHITE_CLASSES

    def is_black(self) -> bool:
        return (self.notenumber % 12) in KEYBOARD_BLACK_CLASSES


class KeyBoard(tk.Frame):
    notebook: ttk.Notebook
    tab1: tk.Frame
    tab2: tk.Frame

    canvas: tk.Canvas
    scrollbar: tk.Scrollbar

    keyboard: tk.Frame
    keylist: list[KeyElement]

    def __init__(self, master: tk.Misc):
        super().__init__(master)
        self.master = master
        self.create_widget()

    def create_widget(self):
        self.notebook = ttk.Notebook(self)
        self.tab1 = tk.Frame(self.notebook)
        self.tab2 = tk.Frame(self.notebook)

        # --- tab1 ---
        self.canvas = tk.Canvas(self.tab1, width=1000, height=72)
        self.keyboard = tk.Frame(self.canvas, width=1750, height=72, bg="#F0F0F0")
        self.scrollbar = tk.Scrollbar(
            self.tab1, orient=tk.HORIZONTAL, command=self.canvas.xview
        )

        self.canvas.create_window((0, 0), window=self.keyboard, anchor=tk.NW)
        self.canvas.configure(xscrollcommand=self.scrollbar.set)

        self.canvas.config(scrollregion=self.canvas.bbox("all"))
        self.keyboard.bind(
            "<Configure>",
            lambda _: self.canvas.configure(scrollregion=self.canvas.bbox("all")),
        )
        self.keyboard.update_idletasks()

        self.keylist = self._create_keylist()

        self.canvas.pack(fill=tk.BOTH, expand=True)
        self.scrollbar.pack(side=tk.BOTTOM, fill=tk.X)

        # --- tab2 ---

        # ------------
        self.notebook.add(self.tab1, text="keyboard")
        self.notebook.add(self.tab2, text="⚙")
        self.notebook.pack(fill=tk.BOTH, expand=True)

    def _create_keylist(self) -> list[KeyElement]:
        x = 0
        keylist: list[KeyElement] = []

        for notenumber in range(0, 128):
            key = KeyElement(self.keyboard, notenumber, width=24, height=72)

            if key.is_white():
                key.place(x=x, y=0, anchor=tk.NW)
                x += 24

            elif key.is_black():
                key.place(x=x, y=0, anchor=tk.N)

            keylist.append(key)

        for key in keylist[:-1]:
            if key.is_black():
                key.lift(aboveThis=keylist[-1])

        return keylist

    def default(self):
        return



import tkinter as tk

import pytest

from src.aulos.ui import KeyBoard, ScaleViewer


@pytest.fixture(scope="function")
def root():
    root = tk.Tk()
    yield root
    root.quit()


def test_Keyboard_create(root):
    keyboard = KeyBoard(root)
    assert isinstance(keyboard, tk.Misc)
    assert keyboard.winfo_exists()


def test_ScaleViewer_create(root):
    scaleviewer = ScaleViewer(root)
    assert isinstance(scaleviewer, tk.Misc)
    assert scaleviewer.winfo_exists()



from ..scale import DiatonicScale, NondiatonicScale
from .note import Key, PitchClass


class Major(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 2, 1, 2, 2, 2, 1],
    key=Key,
    pitchclass=PitchClass,
):
    """Major scale (Ionian mode)

    * Consisting of a specific pattern of whole and half steps: [2, 2, 1, 2, 2, 2, 1].
    * The most commonly used scale in Western music, characterized by a happy or bright sound.
    * Derived from the diatonic system, starting from the root note and ascending.
    * Frequently used as the foundation for other musical scales and modes.
    """


class Minor(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 1, 2, 2, 1, 2, 2],
    key=Key,
    pitchclass=PitchClass,
):
    """Minor scale (Aeolian mode)

    * Consists of the following intervals: [2, 1, 2, 2, 1, 2, 2].
    * Known for its melancholic, somber, or sad sound.
    * The natural minor scale, often used in classical, rock, and pop music.
    * The third degree is lowered by a half-step compared to the major scale.
    * Forms the basis for many modal scales, including the harmonic and melodic minor scales.
    """


class HarmonicMinor(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 1, 2, 2, 1, 3, 1],
    key=Key,
    pitchclass=PitchClass,
):
    """Harmonic minor scale

    * Consists of the following intervals: [2, 1, 2, 2, 1, 3, 1].
    * A variation of the natural minor scale with a raised 7th degree.
    * Known for its exotic, dramatic sound, often used in classical and metal music.
    * The raised 7th degree creates an augmented second between the 6th and 7th degrees.
    * Frequently used to create tension in chord progressions.
    """


class MelodicMinor(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 1, 2, 2, 2, 2, 1],
    key=Key,
    pitchclass=PitchClass,
):
    """Melodic minor scale

    * Consists of the following intervals: [2, 1, 2, 2, 2, 2, 1].
    * In its ascending form, both the 6th and 7th degrees are raised.
    * Known for its smooth, sophisticated sound, often used in jazz and classical music.
    * The descending form typically uses the natural minor scale (Aeolian mode).
    * The raised 6th and 7th degrees help avoid awkward intervals in melodic motion.
    """


class Pentatonic(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [0], [], [0], [0], []],
    base=Major,
    key=Key,
    pitchclass=PitchClass,
):
    """Pentatonic scale

    * A five-note scale derived from the major scale.
    * Composed of the notes: 1, 2, 3, 5, 6 in relation to the major scale.
    * Known for its simplicity and versatility, commonly used in folk, blues, and rock music.
    * The pentatonic scale omits the 4th and 7th degrees of the major scale.
    * Works well in both major and minor contexts, creating a neutral or balanced sound.
    """


class MinorPentatonic(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [], [0], [0], [0], [], [0]],
    base=Minor,
    key=Key,
    pitchclass=PitchClass,
):
    """Minor pentatonic scale

    * A five-note scale derived from the natural minor scale.
    * Composed of the notes: 1, 3, 4, 5, 7 in relation to the natural minor scale.
    * Known for its bluesy, soulful sound, commonly used in rock, blues, and jazz music.
    * Omits the 2nd and 6th degrees of the minor scale.
    * Provides a simple yet expressive framework for improvisation and melody.
    """


class Bluenote(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [-1, 0], [0], [-1, 0], [0], [-1, 0]],
    base=Major,
    key=Key,
    pitchclass=PitchClass,
):
    """Bluenote scale

    * A variation of the pentatonic scale with the addition of flat 5 (blue note).
    * Composed of the notes: 1, 2, 3, 4, 5b, 6, 7.
    * Characterized by its "bluesy" sound, used extensively in jazz, blues, and rock.
    * The flat 5 creates a distinct tension and release, making it effective for expressive improvisation.
    * Often associated with the "blue" feeling, evoking melancholy or sadness.
    """


class Diminish(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [-1], [0, 1], [1], [0], [0]],
    base=Major,
    key=Key,
    pitchclass=PitchClass,
):
    """Diminish scale

    * A symmetrical scale that alternates between whole and half steps.
    * Composed of the intervals: [2, 1, 2, 1, 2, 1, 2].
    * Known for its tension-filled, dissonant sound.
    * Frequently used in jazz and classical music for chromatic and diminished chord progressions.
    * The scale's symmetrical nature creates repeated patterns and interesting melodic possibilities.
    """


class CombDiminish(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [-1], [-1, 0], [1], [0], [0], [-1]],
    base=Major,
    key=Key,
    pitchclass=PitchClass,
):
    """CombDiminish scale

    * A combination of diminished and whole-tone scale patterns.
    * Composed of alternating diminished and whole-step intervals, creating a hybrid structure.
    * Known for its mysterious, unconventional sound, often used in jazz and avant-garde music.
    * The scale's hybrid nature allows for chromatic movement and unusual chord progressions.
    """


class Wholetone(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [0], [1], [1], [1], []],
    base=Major,
    key=Key,
    pitchclass=PitchClass,
):
    """Whole tone scale

    * A scale composed entirely of whole steps.
    * Characterized by its ambiguous, dreamlike sound with no leading tone.
    * The whole tone scale contains only six notes in total.
    * Commonly used in impressionist music, creating a floating, ethereal atmosphere.
    * Its symmetrical structure gives it a unique sound, with no tendency to resolve to a tonic.
    """


class Ionian(
    Major,
    intervals=Major.intervals,
    shift=0,
    key=Key,
    pitchclass=PitchClass,
):
    """Ionian mode (Major scale)

    * Also known as the major scale.
    * A diatonic scale with a happy, bright, or resolved sound.
    * Composed of the intervals: [2, 2, 1, 2, 2, 2, 1].
    * The foundation of Western music harmony and melody.
    * Frequently used in a wide variety of musical genres.
    * This mode starts at the root of the major scale (no shift).
    """


class Dorian(
    Major,
    intervals=Major.intervals,
    shift=1,
    key=Key,
    pitchclass=PitchClass,
):
    """Dorian mode

    * A minor scale with a raised 6th degree.
    * Composed of the intervals: [2, 1, 2, 2, 2, 1, 2].
    * Known for its jazzy, bluesy character, balancing between major and minor.
    * The 6th degree distinguishes Dorian from the natural minor scale.
    * Often used in jazz, blues, and rock for improvisation.
    * This mode starts from the 2nd degree of the major scale.
    """


class Phrygian(
    Major,
    intervals=Major.intervals,
    shift=2,
    key=Key,
    pitchclass=PitchClass,
):
    """Phrygian mode

    * A minor scale with a lowered 2nd degree.
    * Composed of the intervals: [1, 2, 2, 2, 1, 2, 2].
    * Known for its dark, exotic, or Spanish flavor.
    * The lowered 2nd degree gives it a distinct, dissonant sound.
    * Common in flamenco music and jazz.
    * This mode starts from the 3rd degree of the major scale.
    """


class Lydian(
    Major,
    intervals=Major.intervals,
    shift=3,
    key=Key,
    pitchclass=PitchClass,
):
    """Lydian mode

    * A major scale with a raised 4th degree.
    * Composed of the intervals: [2, 2, 2, 1, 2, 2, 1].
    * Known for its bright, dreamy, or floating sound.
    * The raised 4th degree gives it a unique flavor, often used in jazz fusion and progressive rock.
    * This mode starts from the 4th degree of the major scale.
    """


class Mixolydian(
    Major,
    intervals=Major.intervals,
    shift=4,
    key=Key,
    pitchclass=PitchClass,
):
    """Mixolydian mode

    * A major scale with a lowered 7th degree.
    * Composed of the intervals: [2, 2, 1, 2, 2, 1, 2].
    * Known for its bluesy or dominant sound.
    * The lowered 7th degree gives it a less resolved sound compared to the Ionian mode.
    * Often used in rock, blues, and funk.
    * This mode starts from the 5th degree of the major scale.
    """


class Aeorian(
    Major,
    intervals=Major.intervals,
    shift=5,
    key=Key,
    pitchclass=PitchClass,
):
    """Aeolian mode (Natural minor scale)

    * A natural minor scale with no raised or lowered degrees.
    * Composed of the intervals: [2, 1, 2, 2, 1, 2, 2].
    * Known for its melancholic, somber, or sad sound.
    * The foundation for the minor scale in Western music.
    * Often used in classical, rock, and pop music.
    * This mode starts from the 6th degree of the major scale.
    """


class Locrian(
    Major,
    intervals=Major.intervals,
    shift=6,
    key=Key,
    pitchclass=PitchClass,
):
    """Locrian mode

    * A minor scale with a lowered 2nd and 5th degree.
    * Composed of the intervals: [1, 2, 2, 1, 2, 2, 2].
    * Known for its dissonant, unstable, and tense sound.
    * The lowered 5th degree gives it a diminished quality.
    * Rarely used in classical or contemporary music, but found in experimental and avant-garde genres.
    * This mode starts from the 7th degree of the major scale.
    """


class Dorian_f2(
    MelodicMinor,
    intervals=MelodicMinor.intervals,
    shift=1,
    key=Key,
    pitchclass=PitchClass,
):
    """Dorian flat 2 mode (Dorian #2)

    * A modified version of the Dorian mode with a lowered 2nd degree.
    * Composed of the intervals: [1, 2, 2, 2, 2, 1, 2].
    * Known for its exotic, tense sound.
    * This mode is often used in jazz and fusion for a darker minor sound.
    * This mode starts from the 2nd degree of the melodic minor scale.
    """


class Lydian_s5(
    MelodicMinor,
    intervals=MelodicMinor.intervals,
    shift=2,
    key=Key,
    pitchclass=PitchClass,
):
    """Lydian sharp 5 mode (Lydian #5)

    * A Lydian mode with a raised 5th degree.
    * Composed of the intervals: [2, 2, 2, 1, 3, 1, 2].
    * Known for its otherworldly, dreamlike quality.
    * The raised 5th degree creates an augmented chord quality.
    * Often used in jazz and experimental music for rich harmonic tension.
    * This mode starts from the 3rd degree of the melodic minor scale.
    """


class Lydian_f7(
    MelodicMinor,
    intervals=MelodicMinor.intervals,
    shift=3,
    key=Key,
    pitchclass=PitchClass,
):
    """Lydian flat 7 mode (Lydian b7)

    * A Lydian mode with a lowered 7th degree.
    * Composed of the intervals: [2, 2, 2, 2, 1, 2, 1].
    * Known for its mix of major and dominant qualities.
    * Often used in jazz, fusion, and progressive rock.
    * This mode starts from the 4th degree of the melodic minor scale.
    """


class Mixolydian_f6(
    MelodicMinor,
    intervals=MelodicMinor.intervals,
    shift=4,
    key=Key,
    pitchclass=PitchClass,
):
    """Mixolydian flat 6 mode (Mixolydian b6)

    * A Mixolydian mode with a lowered 6th degree.
    * Composed of the intervals: [2, 2, 1, 2, 1, 2, 2].
    * Known for its bluesy, jazzy sound.
    * The lowered 6th degree gives it a minor quality while maintaining the dominant 7th.
    * This mode is commonly used in jazz and fusion.
    * This mode starts from the 5th degree of the melodic minor scale.
    """


class Aeorian_f5(
    MelodicMinor,
    intervals=MelodicMinor.intervals,
    shift=5,
    key=Key,
    pitchclass=PitchClass,
):
    """Aeolian flat 5 mode (Aeolian b5)

    * A natural minor scale with a lowered 5th degree.
    * Composed of the intervals: [2, 1, 2, 1, 2, 2, 2].
    * Known for its dark, exotic sound, often used in jazz and metal.
    * This mode creates a diminished quality by lowering the 5th degree.
    * This mode starts from the 6th degree of the melodic minor scale.
    """


class SuperLocrian(
    MelodicMinor,
    intervals=MelodicMinor.intervals,
    shift=6,
    key=Key,
    pitchclass=PitchClass,
):
    """Super Locrian mode (Altered scale)

    * A scale with altered 2nd, 3rd, 4th, 5th, 6th, and 7th degrees.
    * Composed of the intervals: [1, 2, 1, 2, 1, 2, 2].
    * Known for its highly dissonant, tension-filled sound.
    * Often used in jazz over diminished or dominant chords.
    * The altered scale is a tool for creating complex harmonies.
    * This mode starts from the 7th degree of the melodic minor scale.
    """


class Locrian_n6(
    HarmonicMinor,
    intervals=HarmonicMinor.intervals,
    shift=1,
    key=Key,
    pitchclass=PitchClass,
):
    """Locrian natural 6 mode (Locrian #6)

    * A Locrian mode with a raised 6th degree.
    * Composed of the intervals: [1, 2, 2, 1, 2, 3, 1].
    * Known for its eerie, dissonant sound with a slight major flavor due to the raised 6th degree.
    * This mode is used in jazz and contemporary music for unusual, dissonant harmonic progressions.
    * This mode starts from the 2nd degree of the harmonic minor scale.
    """


class Ionian_s5(
    HarmonicMinor,
    intervals=HarmonicMinor.intervals,
    shift=2,
    key=Key,
    pitchclass=PitchClass,
):
    """Ionian sharp 5 mode (Ionian #5)

    * A major scale with a raised 5th degree.
    * Composed of the intervals: [2, 2, 1, 2, 3, 1, 1].
    * Known for its bright sound with a dramatic twist due to the sharp 5th.
    * Often used in fusion and experimental music to create harmonic tension.
    * This mode starts from the 3rd degree of the harmonic minor scale.
    """


class Dorian_s4(
    HarmonicMinor,
    intervals=HarmonicMinor.intervals,
    shift=3,
    key=Key,
    pitchclass=PitchClass,
):
    """Dorian sharp 4 mode (Dorian #4)

    * A Dorian mode with a raised 4th degree.
    * Composed of the intervals: [2, 1, 2, 3, 2, 1, 2].
    * Known for its tension and dark character.
    * Often used in jazz and fusion for a modern, complex sound.
    * This mode starts from the 4th degree of the harmonic minor scale.
    """


class Mixolydian_f9(
    HarmonicMinor,
    intervals=HarmonicMinor.intervals,
    shift=4,
    key=Key,
    pitchclass=PitchClass,
):
    """Mixolydian flat 9 mode (Phrygian Dominant)

    * A Mixolydian mode with a lowered 9th degree.
    * Composed of the intervals: [1, 2, 2, 2, 1, 2, 2].
    * Known for its exotic, Middle Eastern flavor.
    * Often used in flamenco, metal, and jazz to create tension and drama.
    * This mode starts from the 5th degree of the harmonic minor scale.
    """


class Lydian_s2(
    HarmonicMinor,
    intervals=HarmonicMinor.intervals,
    shift=5,
    key=Key,
    pitchclass=PitchClass,
):
    """Lydian flat 2 mode (Lydian b2)

    * A Lydian mode with a lowered 2nd degree.
    * Composed of the intervals: [1, 3, 2, 1, 2, 2, 2].
    * Known for its dramatic, exotic sound with a raised 4th and lowered 2nd degree.
    * Often used in avant-garde and fusion music to create tension and mystery.
    * This mode starts from the 6th degree of the harmonic minor scale.
    """


class AlteredSuperLocrian(
    HarmonicMinor,
    intervals=HarmonicMinor.intervals,
    shift=6,
    key=Key,
    pitchclass=PitchClass,
):
    """Altered Super Locrian mode (Altered scale with Super Locrian qualities)

    * A highly altered scale that is often used in jazz and fusion.
    * Composed of the intervals: [1, 2, 1, 2, 1, 2, 2].
    * Known for its extreme dissonance and tension, incorporating all possible alterations of the major scale.
    * Features a mix of both diminished and augmented intervals, contributing to its complex and unstable sound.
    * Often used in jazz over dominant chords to create maximum harmonic tension and to resolve to tonic chords.
    * This mode starts from the 7th degree of the harmonic minor scale.
    """



from dataclasses import dataclass

from ..._core import Schema
from ...note.schemas import NoteSchema, PitchClassSchema


@dataclass(frozen=True)
class TunerSchema(Schema):
    reference_notenumber: int
    note: NoteSchema
    pitchclass: PitchClassSchema

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        # [check] reference_notenumber
        if self.reference_notenumber not in self.note.notenumbers:
            raise Exception()



from ..note import BaseKey, BaseNote, BasePitchClass


class PitchClass(
    BasePitchClass,
    intervals=(2, 2, 1, 2, 2, 2, 1),
    symbols_pitchclass=("C", "D", "E", "F", "G", "A", "B"),
    symbols_accidental=("bbb", "bb", "b", "#", "##", "###"),
):
    """ """


class Note(
    BaseNote[PitchClass],
    symbols_notenumber=range(128),
    symbols_octave=(
        "<N>-1",
        "<N>0",
        "<N>1",
        "<N>2",
        "<N>3",
        "<N>4",
        "<N>5",
        "<N>6",
        "<N>7",
        "<N>8",
        "<N>9",
    ),
    reference_notenumber=60,
    reference_octave=5,
    pitchclass=PitchClass,
):
    """ """


class Key(
    BaseKey[PitchClass],
    accidental=1,
    pitchclass=PitchClass,
):
    """ """



"""
12-TET (12 equal temperament)
---

This module provides classes and functions for working with notes, pitch classes,
and scales in the 12-tone equal temperament system.
It includes definitions for various musical scales and modes.
"""

from .note import Key, Note, PitchClass

# Scales
# Modes
from .scale import (
    Aeorian,
    Aeorian_f5,
    AlteredSuperLocrian,
    Bluenote,
    CombDiminish,
    Diminish,
    Dorian,
    Dorian_f2,
    Dorian_s4,
    HarmonicMinor,
    Ionian,
    Ionian_s5,
    Locrian,
    Locrian_n6,
    Lydian,
    Lydian_f7,
    Lydian_s2,
    Lydian_s5,
    Major,
    MelodicMinor,
    Minor,
    MinorPentatonic,
    Mixolydian,
    Mixolydian_f6,
    Mixolydian_f9,
    Pentatonic,
    Phrygian,
    SuperLocrian,
    Wholetone,
)
from .tuner import Equal12Tuner, JustIntonationTuner, MeantoneTuner, PythagoreanTuner

__all__ = [
    # Note classes
    "Key",
    "Note",
    "PitchClass",
    # Modes
    "Aeorian",
    "Aeorian_f5",
    "AlteredSuperLocrian",
    "Dorian",
    "Dorian_f2",
    "Dorian_s4",
    "Ionian",
    "Ionian_s5",
    "Locrian",
    "Locrian_n6",
    "Lydian",
    "Lydian_f7",
    "Lydian_s2",
    "Lydian_s5",
    "Mixolydian",
    "Mixolydian_f6",
    "Mixolydian_f9",
    "Phrygian",
    "SuperLocrian",
    # Scales
    "Bluenote",
    "CombDiminish",
    "Diminish",
    "HarmonicMinor",
    "Major",
    "MelodicMinor",
    "Minor",
    "MinorPentatonic",
    "Pentatonic",
    "Wholetone",
    # tuners
    "JustIntonationTuner",
    "MeantoneTuner",
    "PythagoreanTuner",
    "Equal12Tuner",
]



from .array import index, rotated
from .calculation import wrapped_diff
from .dataclass import from_dict
from .decorators import classproperty

__all__ = [
    "index",
    "rotated",
    "wrapped_diff",
    "classproperty",
    "from_dict",
]



import typing as t
from contextlib import ContextDecorator
from contextvars import ContextVar

from ..setting import Setting


class Context(ContextDecorator):
    setting: t.ClassVar[ContextVar[Setting]] = ContextVar("setting")
    data: t.ClassVar[ContextVar[dict[str, t.Any]]] = ContextVar("data")

    def __init__(
        self,
        setting: Setting,
        **data,
    ) -> None:
        self.__setting = self.setting.set(setting)
        self.__data = self.data.set(data)

    def __enter__(self) -> t.Self:
        return self

    def __exit__(self, *tracebacks):
        self.setting.reset(self.__setting)
        self.data.reset(self.__data)



import typing as t
from typing import ClassVar


class OptimizedMeta(type):
    def __new__(cls, name: str, bases: tuple[type], dct: dict[str, t.Any], **kwargs):
        annotations: dict[str, type] = dct.get("__annotations__", {})
        slots = tuple(
            name
            for name, typ in annotations.items()
            if not (hasattr(typ, "__origin__") and t.get_origin(typ) is ClassVar)
            if not (hasattr(typ, "__origin__") and t.get_origin(typ) is type)
        )
        dct["__slots__"] = slots
        return super(OptimizedMeta, cls).__new__(cls, name, bases, dct, **kwargs)



def compress_ratio(ratio: float) -> float:
    while ratio > 2:
        ratio = ratio / 2
    while ratio < 1:
        ratio = ratio * 2
    return ratio


def standard_tuning_table(p5th_ratio: float) -> tuple[float, ...]:
    # https://en.wikipedia.org/wiki/Just_intonation#Five-limit_tuning
    (ratios := [compress_ratio(p5th_ratio**i) for i in range(12)]).sort()

    return tuple(ratios)


def fivelimit_tuning_table() -> tuple[float, ...]:
    # https://en.wikipedia.org/wiki/Just_intonation#Five-limit_tuning
    (
        ratios := [
            compress_ratio((3**j) * (5**i))
            for i in range(-1, 1 + 1)
            for j in range(-2, 2 + 1)
            if not j == -2
        ]
    ).sort()

    return tuple(ratios)



import tkinter as tk
import tkinter.ttk as ttk
import typing as t

from .... import Scale
from ....TET12 import scale
from ..base import BaseComponent

SCALE_DEFAULTS: tuple[dict[str, type[Scale]], dict[str, type[Scale]]] = (
    {
        scale.Major.__name__: scale.Major,
        scale.Minor.__name__: scale.Minor,
        scale.MelodicMinor.__name__: scale.MelodicMinor,
        scale.HarmonicMinor.__name__: scale.HarmonicMinor,
        scale.Pentatonic.__name__: scale.Pentatonic,
        scale.MinorPentatonic.__name__: scale.MinorPentatonic,
        scale.Diminish.__name__: scale.Diminish,
        scale.CombDiminish.__name__: scale.CombDiminish,
        scale.Wholetone.__name__: scale.Wholetone,
        scale.Bluenote.__name__: scale.Bluenote,
    },
    {
        scale.Aeorian.__name__: scale.Aeorian,
        scale.Aeorian_f5.__name__: scale.Aeorian_f5,
        scale.AlteredSuperLocrian.__name__: scale.AlteredSuperLocrian,
        scale.Dorian.__name__: scale.Dorian,
        scale.Dorian_f2.__name__: scale.Dorian_f2,
        scale.Dorian_s4.__name__: scale.Dorian_s4,
        scale.Ionian.__name__: scale.Ionian,
        scale.Ionian_s5.__name__: scale.Ionian_s5,
        scale.Locrian.__name__: scale.Locrian,
        scale.Locrian_n6.__name__: scale.Locrian_n6,
        scale.Lydian.__name__: scale.Lydian,
        scale.Lydian_f7.__name__: scale.Lydian_f7,
        scale.Lydian_s2.__name__: scale.Lydian_s2,
        scale.Lydian_s5.__name__: scale.Lydian_s5,
        scale.Mixolydian.__name__: scale.Mixolydian,
        scale.Mixolydian_f6.__name__: scale.Mixolydian_f6,
        scale.Mixolydian_f9.__name__: scale.Mixolydian_f9,
        scale.Phrygian.__name__: scale.Phrygian,
        scale.SuperLocrian.__name__: scale.SuperLocrian,
    },
)


class ScaleSelecter(BaseComponent):
    _selected_scalename: tk.StringVar
    _selected_scaleinfo: tk.StringVar

    _scaleselecter_wrap: ttk.Frame
    _scalesetecter_title: ttk.Label
    _scalegroups: list[ttk.Frame]
    _scalebuttons: list[list[ttk.Radiobutton]]

    def __init__(self, master: tk.Misc):
        super().__init__(master)
        self.master = master
        self.create_widget()

    def create_widget(self):
        self._selected_scalename = tk.StringVar()
        self._selected_scaleinfo = tk.StringVar()

        self._scaleselecter_wrap = ttk.Frame(
            self, padding=(24, 8), borderwidth=2, relief=tk.SOLID
        )
        self._scalesetecter_title = ttk.Label(self, text="Scale")
        self._scaleselecter_wrap.pack()
        self._scalesetecter_title.place(relx=0.05, rely=0, anchor=tk.W)

        self._scalegroups = [
            ttk.Frame(self._scaleselecter_wrap, padding=(6, 0))
            for _ in range(len(SCALE_DEFAULTS))
        ]
        self._scalebuttons = [
            [
                ttk.Radiobutton(
                    scalegroup,
                    text=scale,
                    value=scale,
                    variable=self._selected_scalename,
                    command=self._onClickScaleButton,
                )
                for scale in scales.keys()
            ]
            for scalegroup, scales in zip(self._scalegroups, SCALE_DEFAULTS)
        ]

        for scalegroup in self._scalegroups:
            scalegroup.pack(side=tk.LEFT, anchor=tk.NW)

        for scalebuttons in self._scalebuttons:
            for btn in scalebuttons:
                btn.pack(side=tk.TOP, anchor=tk.NW)

    def default(self):
        self._selected_scalename.set(scale.Major.__name__)
        self._onClickScaleButton()

    def _onClickScaleButton(self):
        for scales in SCALE_DEFAULTS:
            name = self._selected_scalename.get()
            if name in scales:
                self._selected_scaleinfo.set(scales[name].__doc__ or "")

        for callback in self.callbacks_onClickScaleButton:
            callback()

    @property
    def scale(self) -> type[Scale] | None:
        for scales in SCALE_DEFAULTS:
            if self.scalename in scales:
                return scales[self.scalename]
        return None

    @property
    def scalename(self) -> str:
        return self._selected_scalename.get()

    @property
    def scaleinfo(self) -> str:
        return self._selected_scaleinfo.get()

    callbacks_onClickScaleButton: list[t.Callable[[], t.Any]]

    def set_callback_onClickScaleButton(self, callback: t.Callable[[], t.Any]):
        if not hasattr(self, "callbacks_onClickScaleButton"):
            self.callbacks_onClickScaleButton = []
        self.callbacks_onClickScaleButton.append(callback)



import typing as t
from dataclasses import dataclass, field
from functools import cached_property
from itertools import chain

from ..._core import Schema
from .pitchclass import PitchClassSchema


@dataclass(frozen=True, slots=True)
class NoteSchema(Schema):
    symbols_notenumber: tuple[int, ...]
    symbols_octave: tuple[str, ...]
    reference_notenumber: int
    reference_octave: int
    pitchclass: PitchClassSchema

    name2number: dict[str, int] = field(init=False)
    number2name: dict[int, tuple[str | None]] = field(init=False)

    def __post_init__(self) -> None:
        self.validate()
        self.initialize()

    def validate(self) -> None:
        # [check] symbols_notenumber
        if not len(self.symbols_notenumber) > 0:
            raise Exception()
        if not all(0 <= v for v in self.symbols_notenumber):
            raise Exception()

        # [check] symbols_octave
        if not len(self.symbols_octave) > 0:
            raise Exception()
        if not all(
            bool(v.find("<N>")) or bool(v.find("<n>")) for v in self.symbols_octave
        ):
            raise Exception()

        # [check] reference_notenumber
        if self.reference_notenumber not in self.symbols_notenumber:
            raise Exception()

        # [check] reference_octave
        if self.reference_octave not in range(len(self.symbols_octave)):
            raise Exception()

    def initialize(self) -> None:
        accidental = len(self.pitchclass.symbols_accidental) // 2
        upper_accidentals = self.pitchclass.symbols_accidental[accidental:]
        lower_accidentals = reversed(self.pitchclass.symbols_accidental[:accidental])

        def create_upper_sequences() -> list[list[str | None]]:
            sequences = []
            for i, acc in enumerate(upper_accidentals, start=1):
                sequence = create_symbol_sequence(suffix=acc)
                for _ in range(i):
                    sequence.insert(0, sequence.pop())
                sequences.append(sequence)
            return sequences

        def create_lower_sequences() -> list[list[str | None]]:
            sequences = []
            for i, acc in enumerate(lower_accidentals, start=1):
                sequence = create_symbol_sequence(suffix=acc)
                for _ in range(i):
                    sequence.append(sequence.pop(0))
                sequences.append(sequence)
            return sequences

        def create_symbol_sequence(
            *, prefix: str = "", suffix: str = ""
        ) -> list[str | None]:
            sequence: list[str | None] = []
            for symbol_octave in self.symbols_octave:
                for deg in range(self.pitchclass.cardinality):
                    if deg in self.pitchclass.positions:
                        index = self.pitchclass.positions.index(deg)
                        pitchname = (
                            prefix + self.pitchclass.symbols_pitchclass[index] + suffix
                        )
                        notename = convert_pitchname_to_notename(
                            pitchname,
                            symbol_octave,
                        )
                        sequence.append(notename)
                    else:
                        sequence.append(None)
            return sequence

        def convert_pitchname_to_notename(pitchname: str, symbol_octave: str) -> str:
            # <N>
            if symbol_octave.find("<N>") >= 0:
                return symbol_octave.replace("<N>", pitchname, 1)
            # <n>
            elif symbol_octave.find("<n>") >= 0:
                return symbol_octave.replace("<n>", pitchname, 1)
            else:
                return symbol_octave + pitchname

        no_accidental_sequence = create_symbol_sequence()
        accidental_upper_sequences = create_upper_sequences()
        accidental_lower_sequences = reversed(create_lower_sequences())
        accidental_sequences = tuple(
            zip(
                *accidental_lower_sequences,
                no_accidental_sequence,
                *accidental_upper_sequences,
            )
        )

        # adjust notenumber
        ref_pitchname = self.pitchclass.symbols_pitchclass[0]
        ref_octave_notename = convert_pitchname_to_notename(
            ref_pitchname, self.symbols_octave[self.reference_octave]
        )
        adjust_notenumber = self.reference_notenumber - no_accidental_sequence.index(
            ref_octave_notename
        )

        name2number = dict(
            chain.from_iterable(
                [
                    [
                        (name, index + adjust_notenumber)
                        for name in names
                        if name is not None
                    ]
                    for index, names in enumerate(accidental_sequences)
                    if index + adjust_notenumber in self.symbols_notenumber
                ]
            )
        )
        number2name = dict(
            [
                (index + adjust_notenumber, name)
                for index, name in enumerate(accidental_sequences)
                if index + adjust_notenumber in self.symbols_notenumber
            ]
        )

        object.__setattr__(self, "name2number", name2number)
        object.__setattr__(self, "number2name", number2name)

    @cached_property
    def notenames(self) -> tuple[str, ...]:
        return tuple(self.name2number.keys())

    @cached_property
    def notenumbers(self) -> tuple[int, ...]:
        return tuple(self.number2name.keys())

    def count_accidental(self, notename: str) -> int:
        self.ensure_valid_notename(notename)
        notenumber = self.convert_notename_to_notenumber(notename)
        notenames = self.convert_notenumber_to_notenames(notenumber)
        return notenames.index(notename) - self.pitchclass.accidental

    def convert_notenumber_to_notename(
        self, notenumber: int, accidental: int
    ) -> str | None:
        self.ensure_valid_notenumber(notenumber)
        return self.number2name[notenumber][self.pitchclass.accidental + accidental]

    def convert_notenumber_to_notenames(
        self, notenumber: int
    ) -> tuple[str | None, ...]:
        self.ensure_valid_notenumber(notenumber)
        return self.number2name[notenumber]

    def convert_notename_to_notenumber(self, notename: str) -> int:
        self.ensure_valid_notename(notename)
        return self.name2number[notename]

    # [unstable]
    def convert_notenumber_to_pitchclass(self, notenumber: int) -> int:
        self.ensure_valid_notenumber(notenumber)
        return notenumber % self.pitchclass.cardinality

    # [unstable]
    def convert_pitchclass_to_notenumber(self, pitchclass: int, octave: int) -> int:
        self.pitchclass.ensure_valid_pitchclass(pitchclass)
        return pitchclass + (self.pitchclass.cardinality * octave)

    # [unstable]
    def convert_notename_to_pitchname(self, notename: str) -> str:
        self.ensure_valid_notename(notename)
        accidental = self.count_accidental(notename)
        notenumber = self.convert_notename_to_notenumber(notename)
        pitchclass = self.convert_notenumber_to_pitchclass(notenumber)
        pitchname = self.pitchclass.convert_pitchclass_to_pitchname(
            pitchclass, accidental
        )
        if pitchname is None:
            raise RuntimeError("unreachable error")
        return pitchname

    # [unstable]
    def convert_pitchname_to_notename(self, pitchname: str, octave: int) -> str:
        self.pitchclass.ensure_valid_pitchname(pitchname)
        accidental = self.pitchclass.count_accidental(pitchname)
        pitchclass = self.pitchclass.convert_pitchname_to_picthclass(pitchname)
        notenumber = self.convert_pitchclass_to_notenumber(pitchclass, octave)
        notename = self.convert_notenumber_to_notename(notenumber, accidental)
        if notename is None:
            raise RuntimeError("unreachable error")
        return notename

    def is_notename(self, value: t.Any) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.notenames

    def is_notenumber(self, value: t.Any) -> t.TypeGuard[int]:
        return isinstance(value, int) and value in self.notenumbers

    def ensure_valid_notename(self, notename: str) -> None:
        if not self.is_notename(notename):
            raise ValueError(
                f"Invalid notename '{notename}'. "
                f"Notename must be a valid musical note name {self.notenames[:3]}."
            )

    def ensure_valid_notenumber(self, notenumber: int) -> None:
        if not self.is_notenumber(notenumber):
            raise ValueError(
                f"Invalid pitchclass '{notenumber}'."
                f"Notenumber must be an integer between {min(self.notenumbers)} and {max(self.notenumbers)} inclusive."
            )



def run_gui():
    import ctypes
    import tkinter as tk

    from aulos.ui import KeyBoard, ScaleViewer

    ctypes.windll.shcore.SetProcessDpiAwareness(1)

    root = tk.Tk()
    root.title("Aulos Application GUI")
    root.geometry("1200x800")
    root.resizable(False, False)
    keyboard = KeyBoard(root)
    keyboard.pack(fill=tk.X)
    scale = ScaleViewer(root)
    scale.pack(fill=tk.X)

    root.mainloop()


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Aulos Application")
    parser.add_argument(
        "-gui", action="store_true", help="run the aulos application in GUI mode"
    )
    args = parser.parse_args()

    if args.gui:
        run_gui()


if __name__ == "__main__":
    main()



import typing as t
from functools import wraps

from .context import Context


class InjectedMeta(type):
    def __new__(cls, name: str, bases: tuple[type], dct: dict[str, t.Any], **kwargs):
        for attr_name, attr_value in dct.items():
            if callable(attr_value):
                if not attr_name == "__init__":
                    continue
                dct[attr_name] = InjectedMeta.inject(attr_value)
        return super(InjectedMeta, cls).__new__(cls, name, bases, dct, **kwargs)

    @staticmethod
    def inject[**P, R](func: t.Callable[P, R]) -> t.Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            injected: dict[str, t.Any] = {}
            if (setting := Context.setting.get(None)) is not None:
                injected["setting"] = setting

            if (data := Context.data.get(None)) is not None:
                injected.update(data)

            injected.update(kwargs)
            return func(*args, **injected)

        return wrapper



import tkinter as tk
import tkinter.ttk as ttk

from .base import BaseComponent
from .selecter import KeySelecter, ScaleSelecter


class _Display(BaseComponent):
    _scale: tk.StringVar
    _keyname: tk.StringVar
    _scalename: tk.StringVar
    _scaleinfo: tk.StringVar

    _wrapper: ttk.Frame
    _scaledisplay: ttk.Label
    _scaleinfodisplay: ttk.Label

    def __init__(self, master: tk.Misc):
        super().__init__(master)
        self.master = master
        self.create_widget()

    def create_widget(self):
        self._scale = tk.StringVar()
        self._keyname = tk.StringVar()
        self._scalename = tk.StringVar()
        self._scaleinfo = tk.StringVar()

        def combine(*args):
            self._scale.set(f"{self.keyname} {self.scalename}")

        self._keyname.trace_add("write", combine)
        self._scalename.trace_add("write", combine)

        self._wrapper = ttk.Frame(self, padding=(24, 8), borderwidth=2, relief=tk.SOLID)
        self._scaledisplay = ttk.Label(
            self._wrapper, textvariable=self._scale, font=("Times", 18)
        )
        self._scaleinfodisplay = ttk.Label(
            self._wrapper, textvariable=self._scaleinfo, font=("Times", 10)
        )

        self._wrapper.pack()
        self._scaledisplay.pack(side=tk.TOP, anchor=tk.NW)
        self._scaleinfodisplay.pack(side=tk.TOP, anchor=tk.NW)

    def default(self):
        return

    @property
    def keyname(self):
        return self._keyname.get()

    @keyname.setter
    def keyname(self, value: str):
        return self._keyname.set(value)

    @property
    def scalename(self):
        return self._scalename.get()

    @scalename.setter
    def scalename(self, value: str):
        self._scalename.set(value)

    @property
    def scaleinfo(self):
        return self._scaleinfo.get()

    @scaleinfo.setter
    def scaleinfo(self, value: str):
        self._scaleinfo.set(value)


class ScaleViewer(BaseComponent):
    scaledisplay: _Display
    keyselecter: KeySelecter
    scaleselecter: ScaleSelecter

    def __init__(self, master: tk.Misc):
        super().__init__(master)
        self.master = master
        self.create_widget()
        self.default()

    def create_widget(self):
        self.scaledisplay = _Display(self)
        self.keyselecter = KeySelecter(self)
        self.scaleselecter = ScaleSelecter(self)

        self.keyselecter.set_callback_onClickKeyButton(self.display_scaledisplay)
        self.scaleselecter.set_callback_onClickScaleButton(self.display_scaledisplay)

        self.scaledisplay.pack(side=tk.TOP, anchor=tk.W, expand=True)
        self.keyselecter.pack(side=tk.LEFT, anchor=tk.N)
        self.scaleselecter.pack(side=tk.LEFT, anchor=tk.N)

    def default(self):
        self.keyselecter.default()
        self.scaleselecter.default()

    def display_scaledisplay(self):
        self.scaledisplay.keyname = self.keyselecter.keyname
        self.scaledisplay.scalename = self.scaleselecter.scalename
        self.scaledisplay.scaleinfo = self.scaleselecter.scaleinfo



import typing as t
from typing import TYPE_CHECKING

from .._core import AulosObject
from .._core.utils import index
from .schemas import PitchClassSchema

if TYPE_CHECKING:
    from ..scale import Scale  # pragma: no cover


class BasePitchClass(AulosObject[PitchClassSchema]):
    _pitchclass: int
    _pitchnames: tuple[str | None, ...]
    _pitchname: str | None
    _scale: t.Optional["Scale"]

    def __init__(
        self,
        identify: int | str | t.Self,
        *,
        scale: t.Optional["Scale"] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, BasePitchClass):
            self._pitchclass = identify._pitchclass
            self._pitchnames = identify._pitchnames
            self._pitchname = identify._pitchname
            self._scale = None
            self.scale = identify._scale

        elif self.is_pitchclass(identify):
            pitchnames = self.schema.convert_pitchclass_to_pitchnames(identify)
            self._pitchclass = identify
            self._pitchnames = pitchnames
            self._pitchname = None
            self._scale = None
            self.scale = scale

        elif self.is_pitchname(identify):
            pitchclass = self.schema.convert_pitchname_to_picthclass(identify)
            pitchnames = self.schema.convert_pitchclass_to_pitchnames(pitchclass)
            self._pitchclass = pitchclass
            self._pitchnames = pitchnames
            self._pitchname = identify
            self._scale = None
            self.scale = scale

        else:
            raise ValueError()

    def __init_subclass__(
        cls,
        *,
        intervals: t.Sequence[int],
        symbols_pitchclass: t.Sequence[str],
        symbols_accidental: t.Sequence[str],
        **kwargs,
    ) -> None:
        schema = PitchClassSchema(
            tuple(intervals),
            tuple(symbols_pitchclass),
            tuple(symbols_accidental),
        )
        super().__init_subclass__(schema=schema, **kwargs)

    @property
    def pitchclass(self) -> int:
        return self._pitchclass

    @property
    def pitchnames(self) -> list[str]:
        return [n for n in self._pitchnames if n is not None]

    @property
    def pitchname(self) -> str | None:
        return self._pitchname

    @pitchname.setter
    def pitchname(self, name: str):
        if self.is_pitchname(name) and name in self._pitchnames:
            self._pitchname = name

    @property
    def scale(self) -> t.Optional["Scale"]:
        return self._scale

    @scale.setter
    def scale(self, scale: t.Optional["Scale"]):
        from ..scale import Scale

        if isinstance(scale, Scale):
            self._scale = scale
            pitchclass = (int(self) - int(scale.key)) % self.schema.cardinality

            if (idx := index(scale.positions, pitchclass)) is not None:
                self._pitchname = self.schema.convert_pitchclass_to_pitchname(
                    self._pitchclass, scale.signatures[idx]
                )

    @classmethod
    def is_pitchname(cls, pitchname: t.Any) -> t.TypeGuard[str]:
        return cls.schema.is_pitchname(pitchname)

    @classmethod
    def is_pitchclass(cls, pitchclass: t.Any) -> t.TypeGuard[int]:
        return cls.schema.is_pitchclass(pitchclass)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, t.SupportsInt):
            return NotImplemented
        return int(self) == int(other)

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __add__(self, other: t.SupportsInt) -> t.Self:
        pitchclass = (int(self) + int(other)) % self.schema.cardinality
        return self.__class__(pitchclass, scale=self.scale, setting=self.setting)

    def __sub__(self, other: t.SupportsInt) -> t.Self:
        pitchclass = (int(self) - int(other)) % self.schema.cardinality
        return self.__class__(pitchclass, scale=self.scale, setting=self.setting)

    def __int__(self) -> int:
        return self._pitchclass

    def __str__(self) -> str:
        return f"<PitchClass: {self.pitchname or self.pitchnames}, scale: {self.scale}>"

    def __repr__(self) -> str:
        return f"<PitchClass: {self.pitchname or self.pitchnames}, scale: {self.scale}>"



from dataclasses import dataclass
from itertools import starmap

from ..._core import Schema
from ...note.schemas import PitchClassSchema


@dataclass(frozen=True, slots=True)
class ScaleSchema(Schema):
    pitchclass: PitchClassSchema

    def generate_scale_signatures(self, intervals: tuple[int, ...]) -> tuple[int, ...]:
        diff = list(
            starmap(lambda x, y: y - x, zip(self.pitchclass.intervals, intervals))
        )
        signature = []
        for i in range(len(self.pitchclass.intervals)):
            cur = i % len(self.pitchclass.intervals)
            next = (i + 1) % len(self.pitchclass.intervals)
            signature.append(diff[cur])
            diff[next] = diff[next] + diff[cur]
            diff[cur] = 0
        return tuple(signature[-1:] + signature[:-1])



import typing as t


class classproperty[T]:
    def __init__(self, method: t.Callable[..., T] | None = None):
        self.fget = method

    def __get__(self, _, cls=None) -> T:
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget(cls)

    def getter(self, method: t.Callable[..., T]):
        self.fget = method
        return self



def wrapped_diff(lhs: int, rhs: int, max: int | None = None) -> int:
    if max is not None:
        result1 = lhs - rhs
        result2 = (lhs + (max if lhs < rhs else 0)) - (rhs + (max if lhs > rhs else 0))
        return result1 if abs(result1) < abs(result2) else result2
    return lhs - rhs



"""Scale
---
"""

from . import schemas
from .diatonic import DiatonicScale, NondiatonicScale
from .scale import Scale

__all__ = [
    "Scale",
    "DiatonicScale",
    "NondiatonicScale",
    "schemas",
]



import typing as t
from dataclasses import dataclass
from functools import cached_property

from ..._core import Schema
from ..._core.utils import wrapped_diff
from .pitchclass import PitchClassSchema


@dataclass(frozen=True, slots=True)
class KeySchema(Schema):
    accidental: int
    pitchclass: PitchClassSchema

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        # [check] accidental
        if not self.accidental > 0:
            raise Exception()
        if not self.accidental < self.pitchclass.accidental:
            raise Exception()

    @cached_property
    def keynames(self) -> tuple[str, ...]:
        keynames = [
            pitchname
            for pitchname in self.pitchclass.pitchnames
            if abs(self.pitchclass.count_accidental(pitchname)) <= self.accidental
        ]
        return tuple(keynames)

    def generate_key_signatures(self, keyname: str) -> tuple[int, ...]:
        self.ensure_valid_keyname(keyname)
        positions = []
        r_symbol = self.pitchclass.convert_pitchname_to_symbol(keyname)
        r_pitchclass = self.pitchclass.convert_pitchname_to_picthclass(keyname)

        idx = self.pitchclass.symbols_pitchclass.index(r_symbol)
        symbols = (
            self.pitchclass.symbols_pitchclass[idx:]
            + self.pitchclass.symbols_pitchclass[:idx]
        )

        for pos, symbol in zip(self.pitchclass.positions, symbols):
            n_pos = self.pitchclass.convert_pitchname_to_picthclass(symbol)
            a_pos = (r_pitchclass + pos) % self.pitchclass.cardinality
            positions.append(wrapped_diff(a_pos, n_pos, self.pitchclass.cardinality))
        return tuple(positions)

    def is_keyname(self, value: t.Any) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.keynames

    def ensure_valid_keyname(self, keyname: str) -> None:
        if not self.is_keyname(keyname):
            raise ValueError()
